package bzr;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;

import mgui.command.CommandInstance;
import mgui.io.util.IoFunctions;

/**************************************************************
 * Functions for Bazaar issues
 * 
 * @author Andrew
 *
 */
public class BzrFunctions extends CommandInstance {

	public static void main(String[] args) throws Exception{
		
		BzrFunctions instance = new BzrFunctions();
		instance.execute(args);
		
	}
	
	@Override
	protected boolean run_command(String command) {
		
		if (command.equals("undo-revert")){
			return undo_revert();
		}
		
		return false;
		
	}

	
	/********************************************************************
	 * Bazaar's "revert" command will delete new uncommitted code (and apparently has an issue
	 * committing code properly, so never use it if at all possible). However, it will fortunately 
	 * save the annihilated code in a new folder, with subfolders and the uncommitted changes,
	 * with names appended with a "~1~" character (or ~i~ where i is the i_th time the file was
	 * reverted). This function simply copies these files back to where they should be. 
	 * 
	 * @return
	 */
	boolean undo_revert(){
		
		boolean fail_on_error = parameters.containsKey("fail_on_error");
		
		String revert_suffix = parameters.get("revert_suffix");
		if (revert_suffix == null)
			revert_suffix = ".~1~";
		
		String output_suffix = parameters.get("output_suffix");
		if (output_suffix == null) output_suffix = "";
		
		String s = parameters.get("revert_dir");
		File revert_dir = new File(s);
		if (!revert_dir.exists() || !revert_dir.isDirectory()){
			log("undo_revert: revert_dir does not exist or is not a directory.");
			return false;
			}
		
		File output_dir = new File(s.substring(0, s.lastIndexOf(revert_suffix)) + output_suffix);
		if (!output_dir.exists() && !output_dir.mkdir()){
			log("undo_revert: Cannot create output directory.");
			return false;
			}
		
		boolean success = revert_bzr_dir(revert_dir, output_dir, fail_on_error);
		
		if (success)
			log("undo-revert: Success.");
		else
			log("undo-revert: Finished with errors.");
		
		return success;
		
	}
	
	boolean revert_bzr_dir(File revert, File output, boolean fail_on_error) {
		
		//recursively copy, removing suffix
		
		String[] java_files = revert.list(new FilenameFilter(){
			@Override
			public boolean accept(File dir, String name) {
				return name.endsWith(".java");
			}
		});
		
		String[] subdirs = revert.list(new FilenameFilter(){
			@Override
			public boolean accept(File dir, String name) {
				File path = new File(dir.getAbsolutePath() + File.separator + name);
				return path.isDirectory();
			}
		});
		
		// Java files
		boolean success = true;
		int success_count = 0, fail_count = 0;
		
		for (int i = 0; i < java_files.length; i++){
			File r_file = new File(revert.getAbsolutePath() + File.separator + java_files[i]);
			File o_file = new File(output.getAbsolutePath() + File.separator + java_files[i]);
			try{
				IoFunctions.copyFile(r_file, o_file);
				log("undo-revert: Copied '" + r_file.getAbsolutePath() + "'"); 
				log("undo-revert:     to '" + o_file.getAbsolutePath() + "'");
				success_count++;
			}catch (IOException ex){
				log("undo-revert: Could not copy file '" + r_file + "'\nDetails:" + ex.getMessage());
				if (fail_on_error)
					return false;
				success = false;
				fail_count++;
				}
			}
		
		if (java_files.length > 0)
			log("undo-revert: Copied from folder '" + revert.getName() + "': " + success_count + " successful, " +
							  fail_count + " failures.");
		
		// Sub directories
		for (int i = 0; i < subdirs.length; i++){
			String subdir = subdirs[i];
			File revert_dir = new File(revert.getAbsolutePath() + File.separator + subdir);
			File output_dir = new File(output.getAbsolutePath() + File.separator + 
									   subdir.substring(0, subdir.lastIndexOf(".~1~")));
			if (!output_dir.exists() && !output_dir.mkdir()){
				log("undo_revert: Cannot create output directory '" + output_dir.getAbsolutePath() + "'.");
				if (fail_on_error)
					return false;
				fail_count++;
			}else{
				success &= revert_bzr_dir(revert_dir, output_dir, fail_on_error);
				}
			
			}
		
		return success;
		
	}
	
	
}
