package mgui.command;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

import mgui.geometry.Grid3D;
import mgui.geometry.volume.VolumeFunctions;
import mgui.interfaces.shapes.Volume3DInt;
import mgui.io.standard.nifti.NiftiVolumeLoader;
import mgui.io.standard.nifti.NiftiVolumeWriter;
import mgui.io.util.IoFunctions;

public class ImagingFunctions extends CommandInstance {

	public static void main(String[] args) throws Exception{
		
		ImagingFunctions instance = new ImagingFunctions();
		instance.execute(args);
		
	}
	
	@Override
	public boolean run_command(String command){
		
		if (command.equals("calculate-average-volume")){
			return calculate_average_volume();
			}
		
		return false;
		
	}
	
	/*********************************************
	 * Calculates an average volume from a list of Nifti-format volumes. 
	 * 
	 * <p>Parameters [required]
	 * <ul>
	 * <li>instance_dir			The instance directory
	 * <li>input_file_pattern	Pattern specifying a path to each input volume file, relative to each instance 
	 * 							subdirectory. Use "{subject}" to insert the current instance name into 
	 * 							the path.
	 * <li>output_file			The output volume file
	 * </ul>
	 * 
	 * <p>Parameters [optional]
	 * <ul>
	 * <li>instance_list_file	A text file listing the instances to average ("\n"-delimited); no file
	 * 							indicates that all instances are to be used
	 * <li>log_file				File to place log comments and errors; no log_file specifies that comments
	 * 							are to be sent to the console
	 * <li>nifti_format			The data format for the output volume; one of double, float, int, short, and byte.
	 * 							The default is int.
	 * </ul>
	 */
	protected boolean calculate_average_volume(){
		
		// Get subdirectories
		String param = parameters.get("instance_dir");
		if (param == null){
			log("calculate_average_volume: Parameter 'instance_dir' is required.");
			return false;
			}
		
		File instance_dir = new File(param);
		
		if (!instance_dir.exists() || !instance_dir.isDirectory()){
			log("calculate_average_volume: Directory '" + instance_dir.getAbsolutePath() + "' does not " +
														"exist or is not a directory.");
			return false;
			}
		
		String input_pattern = parameters.get("input_file_pattern");
		if (input_pattern == null){
			log("calculate_average_volume: Parameter 'input_file_pattern' is required.");
			return false;
			}
		
		input_pattern = input_pattern.replace("/", File.separator);
		input_pattern = input_pattern.replace("\\", File.separator);
		input_pattern = param + File.separator + input_pattern;
		
		param = parameters.get("output_file");
		if (param == null){
			log("calculate_average_volume: Parameter 'output_file' is required.");
			return false;
			}
		
		File output_file = new File(param);
		
		try{
			if (!output_file.exists() && !output_file.createNewFile()){
				log("calculate_average_volume: File '" + output_file.getAbsolutePath() + "' could not be created.");
				return false;
				}
		}catch (IOException e){
			log("calculate_average_volume: File '" + output_file.getAbsolutePath() + "' could not be created.");
			return false;
			}
		
		String number_format = parameters.get("nifti_format");
		if (number_format == null){
			number_format = "int";
			}
		
		ArrayList<String> instances = new ArrayList<String>();
		
		// Load subject list
		param = parameters.get("instance_list_file");
		if (param == null){
			log("calculate_average_volume: No instance file specified; using all instances.");
			String[] list = instance_dir.list(IoFunctions.getDirFilter());
			for (int i = 0; i < list.length; i++)
				instances.add(list[i]);
		}else{
			File list_file = new File(param);
			log("calculate_average_volume: Loading instances from '" + list_file.getAbsolutePath() + "'.");
			if (!list_file.exists()){
				log("calculate_average_volume: File '" + list_file.getAbsolutePath() + "' not found.");
				return false;
				}
			try{
				BufferedReader reader = new BufferedReader(new FileReader(list_file));
				String line = reader.readLine();
				while (line != null){
					instances.add(line);
					line = reader.readLine();
					}
				reader.close();
					
			}catch (IOException e){
				log("calculate_average_volume: Error reading file '" + list_file.getAbsolutePath() + "'.");
				return false;
				}
			}
		
		if (instances.size() < 1){
			log("calculate_average_volume: No instances found!");
			return false;
			}
		
		// For each instance, read volume and add to sum 
		int success_count = 0, fail_count = 0;
		Grid3D average_grid = null;
		
		for (int i = 0; i < instances.size(); i++){
			String instance = instances.get(i);
			
			try{
				String instance_path = input_pattern.replace("{subject}", instance);
				File input_file = new File(instance_path);
				NiftiVolumeLoader loader = new NiftiVolumeLoader(input_file);
				Volume3DInt volume = loader.loadVolume(null);
				
				if (volume == null){
					fail_count++;
				}else{
					success_count++;
					if (i == 0){
						average_grid = volume.getGrid();
					}else{
						average_grid = VolumeFunctions.performOperationBlocking(average_grid, 
																				volume.getGrid(), 
																				VolumeFunctions.Operation.Add, 
																				null);
						}
					log("calculate_average_volume: added instance '" + instance + "'.");
					}
				
			}catch (Exception e){
				fail_count++;
				log("calculate_average_volume: error adding instance '" + instance + "'.");
				}
			
			}
		
		if (success_count == 0){
			log("calculate_average_volume: All instances failed...");
			return false;
			}
		
		VolumeFunctions.performOperationBlocking(average_grid, 
												 (double)success_count, 
												 VolumeFunctions.Operation.Divide, 
												 null); 
		
		try{
			NiftiVolumeWriter writer = new NiftiVolumeWriter(output_file);
			writer.writeVolume(new Volume3DInt(average_grid));
		}catch (IOException e){
			log("calculate_average_volume: Error writing output to '" + output_file.getAbsolutePath() + "'.");
			return false;
			}
		
		log("calculate_average_volume: Done; " + success_count + " successes and " + fail_count + " failures.");
		log("calculate_average_volume: Output written to '" + output_file.getAbsolutePath() + "'.");
		
		return fail_count == 0;
		
	}
	
}
