package mgui.command;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

import javax.swing.JFrame;

import mgui.geometry.Grid3D;
import mgui.geometry.mesh.MeshEngine;
import mgui.interfaces.attributes.AttributeList;
import mgui.interfaces.graphics.InterfaceGraphic3D;
import mgui.interfaces.shapes.Mesh3DInt;
import mgui.interfaces.shapes.ShapeModel3D;
import mgui.interfaces.shapes.ShapeSet3DInt;
import mgui.interfaces.shapes.Volume3DInt;
import mgui.io.domestic.maps.ContinuousColourMapLoader;
import mgui.io.domestic.shapes.VolumeInputOptions;
import mgui.io.foreign.minc.MincSurfaceLoader;
import mgui.io.standard.nifti.NiftiVolumeLoader;
import mgui.numbers.MguiDouble;
import mgui.numbers.MguiNumber;
import mgui.stats.StatFunctions;

/******************************************************
 * Various functions operating on mgui shapes.
 * 
 * 
 * @author Andrew Reid
 * @version 1.0
 * @since 1.0
 *
 */
public class ShapeFunctions extends CommandInstance {

	public static void main(String[] args) throws Exception{
		ShapeFunctions instance = new ShapeFunctions();
		instance.execute(args);
	}
	
	
	@Override
	protected boolean run_command(String command) {
		
		if (command.equals("volume-to-surface-parameters")){
			return volume_to_surface_parameters();
			}
		
		return false;
	}
	
	/*********************************************
	 * Projects data using a Gaussian normal/tangent field, from a surface to a volume and back, across a range
	 * of <code>sigma_normal</code> and <code>sigma_tangent</code> parameters. Iterates over <code>iterations</code>
	 * number of random vertex data sets, and reports the error as mean RMSE for each parameter combination. 
	 * 
	 * <p>Parameters [required]
	 * <ul>
	 * <li>volume_file			The Nifti format volume file containing data to be projected
	 * <li>surface_file			The Civet format surface mesh file
	 * <li>output_file			The output text file, containing tab-delimited RMSE
	 * <li>sigma_normal_min		The minimal sigma_normal value
	 * <li>sigma_normal_max		The maximal sigma_normal value
	 * <li>sigma_normal_step	The step value for sigma_normal at each iteration
	 * <li>sigma_tangent_min	The minimal sigma_tangent value
	 * <li>sigma_tangent_max	The maximal sigma_tangent value
	 * <li>sigma_tangent_step	The step value for sigma_tangent at each iteration
	 * <li>iterations			Number of iterations for random vertex data
	 * </ul>
	 * 
	 * <p>Parameters [optional]
	 * <ul>
	 * <li>output_directory		If specified, the projected data (surface and volume) will be written to this 
	 * 							directory
	 * <li>log_file				File to place log comments and errors; no log_file specifies that comments
	 * 							are to be sent to the console
	 * <li>nifti_format			The data format for the output volume; one of double, float, int, short, and byte.
	 * 							The default is int.
	 * <li>show_shapes			Flag which, if true, displays a 3D window with the volume and surface in it
	 * <li>distance_normal_max	The maximum distance at which to project between vertex and voxel in the normal
	 * 							direction; default is 2.0.
	 * <li>distance_tangent_max	The maximum distance at which to project between vertex and voxel in the tangent
	 * 							direction; default is 2.0.
	 * </ul>
	 */
	public boolean volume_to_surface_parameters(){
		
		boolean show_shapes = parameters.containsKey("show_shapes");
		
		String param = parameters.get("volume_file");
		if (param == null){
			log("volume_to_surface_parameters: Parameter 'volume_file' is required.");
			return false;
			}
		
		File volume_file = new File(param);
		
		if (!volume_file.exists()){
			log("volume_to_surface_parameters: Cannot find volume file '" + volume_file.getAbsolutePath() +"'.");
			return false;
			}
		
		param = parameters.get("iterations");
		if (param == null){
			log("volume_to_surface_parameters: Parameter 'iterations' is required.");
			return false;
			}
		
		int iterations = Integer.valueOf(param);
		
		param = parameters.get("surface_file");
		if (param == null){
			log("volume_to_surface_parameters: Parameter 'surface_file' is required.");
			return false;
			}
		
		File surface_file = new File(param);
		
		if (!surface_file.exists()){
			log("volume_to_surface_parameters: Cannot find surface file '" + surface_file.getAbsolutePath() +"'.");
			return false;
			}
		
		param = parameters.get("output_file");
		if (param == null){
			log("volume_to_surface_parameters: Parameter 'output_file' is required.");
			return false;
			}
		
		File output_file = new File(param);
		try{
		if (!output_file.exists() && !output_file.createNewFile()){
			log("volume_to_surface_parameters: Cannot create output_file '" + output_file.getAbsolutePath() + ".");
			return false;
			}
		}catch (IOException e){
			log("volume_to_surface_parameters: Cannot create output_file '" + output_file.getAbsolutePath() + ".");
			return false;
			}
		
		param = parameters.get("sigma_normal_min");
		if (param == null){
			log("volume_to_surface_parameters: Parameter 'sigma_normal_min' is required.");
			return false;
			}
		double sigma_normal_min = Double.valueOf(param);
		param = parameters.get("sigma_normal_max");
		if (param == null){
			log("volume_to_surface_parameters: Parameter 'sigma_normal_max' is required.");
			return false;
			}
		double sigma_normal_max = Double.valueOf(param);
		param = parameters.get("sigma_normal_step");
		if (param == null){
			log("volume_to_surface_parameters: Parameter 'sigma_normal_step' is required.");
			return false;
			}
		double sigma_normal_step = Double.valueOf(param);
		
		param = parameters.get("sigma_tangent_min");
		if (param == null){
			log("volume_to_surface_parameters: Parameter 'sigma_tangent_min' is required.");
			return false;
			}
		double sigma_tangent_min = Double.valueOf(param);
		param = parameters.get("sigma_tangent_max");
		if (param == null){
			log("volume_to_surface_parameters: Parameter 'sigma_tangent_max' is required.");
			return false;
			}
		double sigma_tangent_max = Double.valueOf(param);
		param = parameters.get("sigma_tangent_step");
		if (param == null){
			log("volume_to_surface_parameters: Parameter 'sigma_tangent_step' is required.");
			return false;
			}
		double sigma_tangent_step = Double.valueOf(param);
		
		double distance_normal_max = 2.0;
		param = parameters.get("distance_normal_max");
		if (param != null){
			distance_normal_max = Double.valueOf(param);
			}
		
		double distance_tangent_max = 2.0;
		param = parameters.get("distance_tangent_max");
		if (param != null){
			distance_tangent_max = Double.valueOf(param);
			}
		
		//Load volume and surface
		try{
			NiftiVolumeLoader loader = new NiftiVolumeLoader(volume_file);
			VolumeInputOptions options = new VolumeInputOptions();
			options.set_origin = true;
			options.origin_x = -95;
			options.origin_y = -130;
			options.origin_z = -80;
			
			ContinuousColourMapLoader c_loader = new ContinuousColourMapLoader(new File(
					"C:\\Documents and Settings\\atreid\\My Documents\\Java\\eclipse\\workspace\\mgui-core\\src\\cmaps\\minc_grey.cmap"));
			
			options.colour_map = c_loader.loadMap();
			options.colour_map.mapMin = 500;
			options.colour_map.mapMax = 700;
			options.has_alpha = true;
			options.window_mid = 0.5855;
			options.window_width = 3.628;
			options.scale = 0.00087;
			options.intercept = -1;
			
			Grid3D grid = loader.getGrid3D(options);
			Volume3DInt volume = new Volume3DInt(grid);
			
			MincSurfaceLoader s_loader = new MincSurfaceLoader(surface_file);
			Mesh3DInt mesh_int = s_loader.loadSurface();
						
			MeshEngine engine = new MeshEngine();
			AttributeList volume_attributes = engine.getAttributes("Map From Volume (Gaussian)");
			AttributeList surface_attributes = engine.getAttributes("Map To Volume (Gaussian)");
			
			surface_attributes.setValue("mesh_column", "projected");
			
			if (show_shapes){
				JFrame frame = new JFrame();
				InterfaceGraphic3D window = new InterfaceGraphic3D();
				
				ShapeSet3DInt set = new ShapeSet3DInt();
				set.addShape(volume);
				set.addShape(mesh_int);
				
				ShapeModel3D model = new ShapeModel3D("model", set);
				
				window.setSource(model);
				frame.setContentPane(window);
				
				frame.setVisible(true);
				}
				
			BufferedWriter writer = new BufferedWriter(new FileWriter(output_file));
			writer.write("Sigmas");
			for (double sigma_t = sigma_tangent_min; sigma_t < sigma_tangent_max; sigma_t += sigma_tangent_step)
				writer.write("\t" + sigma_t);
			
			HashMap<Integer, ArrayList<MguiNumber>> vertex_data = new HashMap<Integer, ArrayList<MguiNumber>>();
			
						
			for (double sigma_n = sigma_normal_min; sigma_n < sigma_normal_max; sigma_n += sigma_normal_step){
				writer.write("\n" + sigma_n);
				log("Starting sigma_n = " + sigma_n);
				for (double sigma_t = sigma_tangent_min; sigma_t < sigma_tangent_max; sigma_t += sigma_tangent_step){
					
					log("Starting sigma_t = " + sigma_t);
					double rmse = 0;
					
					for (int i = 0; i < iterations; i++){
						
						log("Iteration " + i);
						
						//Random surface data
						ArrayList<MguiNumber> data = vertex_data.get(i);
						
						if (data == null){
							int n = mesh_int.getMesh().n;
							data = new ArrayList<MguiNumber>(n);
							for (int j = 0; j < n; j++){
								data.add(new MguiDouble(Math.random()));
								}
							}
						
						mesh_int.addVertexData("projected", data);
						
						//First, surface to volume
						surface_attributes.setValue("sigma_normal", new MguiDouble(sigma_n));
						surface_attributes.setValue("sigma_tangent", new MguiDouble(sigma_t));
						surface_attributes.setValue("sigma_max_normal", new MguiDouble(distance_normal_max));
						surface_attributes.setValue("sigma_max_tangent", new MguiDouble(distance_tangent_max));
						
						Grid3D new_grid = engine.mapMeshToGrid3D(mesh_int, grid, "Map To Volume (Gaussian)", null);
						//Volume3DInt new_volume = new Volume3DInt(new_grid);
						
						//First, volume to surface
						volume_attributes.setValue("Sigma_normal", new MguiDouble(sigma_n));
						volume_attributes.setValue("Sigma_tangent", new MguiDouble(sigma_t));
						surface_attributes.setValue("Sigma_max_normal", new MguiDouble(distance_normal_max));
						surface_attributes.setValue("Sigma_max_tangent", new MguiDouble(distance_tangent_max));
						
						ArrayList<MguiNumber> new_data = engine.mapGrid3DToMesh(mesh_int.getMesh(), new_grid, "Map From Volume (Gaussian)");
						
						rmse += StatFunctions.getRootMeanSquaredError(data, new_data); 
						
						}
					
					writer.write("\t" + (rmse / (double)iterations));
					
					}
				
				}
			
			writer.close();
			return true;
		
		}catch (Exception e){
			log("volume_to_surface_parameters: Exception encountered:\n" + e.getLocalizedMessage());
			return false;
			}
		
		
	}
	

}
